//........................................................
// FreqGenNEU10b
//
// Hauptschleife fuer Prototypen-Projekt zur Steuerung 
// des Si5351 als Oszillator für QRP-Projekte
//
// Code released under the terms of the GNU GPL 3.0 license
//
// Axel Schultze, DK4AQ
// Stand: 03.10.2015
//........................................................  

// History:
// mit Umschaltung des Frequenz-Inkrements ueber Drehgeber-Knopf
// Abspeicherung der Start-Konfiguration im EEPROM ueber Taster (deaktiviert)
// Einlesen einer Spannung und Pegelanzeige.
// Anbindung des SI531
// EE funktioniert
// Inkrement erweitert
// Pegelanzeige umgestellt auf Feinstrich (eigene Char. für Strichkombinationen)

// 20-06-2015: Modularisierung, Umbenennung Konstanten / Variablen
// 29-06-2015: Inkrementanzahl und Darstellung erweitert
// 09-07-2015: Funktiomiert mit optischem Encoder (Grayhill), 
//             anderes Signalverhalten !
// 06-08-2015: Mech. Prellungen mit Kurbelknopf von SEGOR OK (Massentraegheit)
// 15-09-2015: Fehlerbehebung nach Modularisierung
// 26-09-2015: Fehlerbehebung nach Erstellung der Struktogramme
// 26-09-2015: Feature Auf-/Abrundung der Frequenz bei erster Frequenzverstellung 
//             auf Schrittweite
// 01-10-2015: Fehlerkorrektur aus Struktogramm/Dokumentation
// 02.10.2015: Pegelanzeige optimiert, CRC-Sicherumng für EEPROM-Ablage
//******************************************************************************
#include <LiquidCrystal.h>
#include <EEPROM.h>
#include <Wire.h>
#include <stdint.h>
#include "Arduino.h"
#include "si5351.h"
Si5351 si5351;

//******************************************************************************
// Frequenzgrenzen in Hz
#define FRE_MIN 1600000L
#define FRE_MAX 30000000L
#define FRE_START 3500000L
//deltaF in Hz
#define FRE_INCREM_A 1L
#define FRE_INCREM_B 10L
#define FRE_INCREM_C 100L
#define FRE_INCREM_D 1000L
#define FRE_INCREM_E 10000L

//Inkrementstufen ueber dre_StateIncr indexiert
unsigned long fre_increm[6] = {
  1, FRE_INCREM_A, FRE_INCREM_B, FRE_INCREM_C, FRE_INCREM_D, FRE_INCREM_E};
boolean fre_increm_neu = LOW; //Merker Inkrement wurde umgeschaltet
#define EE_KONFIG_START 0 // Startadresse der Konfig-Struktur im EEPROM

//Frequenz
unsigned long fre_frequenz = 0;  // Variable Frequenz
unsigned long fre_frequenz_alt = 0;  // Variable Frequenz des letzten Durchlaufs

boolean fre_new = HIGH;  // Merker Frequenz wurde verändert
unsigned long fre_deltaF = fre_increm[1]; // Frequenzinkrement
unsigned char dis_ccc = 0; //Anzahl ausgegebener Char bei lcd.print Freq.
unsigned long fre_max = FRE_MAX;// obere Frequenzgrenze
unsigned long fre_min = FRE_MIN;// untere Frequenzgrenze
//******************************************************************************
// Drehgeber
#define DRE_UP HIGH
#define DRE_DN LOW
int dre_IntPin2 = 2; //Pin 2 Ausgang A
int dre_IntPin3 = 3; //Pin 3 Ausgang B
unsigned char dre_value[4] = {0,1,3,2}; // Ablage Umcodierung Gray / binär
int dre_Tast = 4; // Pin 4 als Eingang fuer Drehgeber-Taster
boolean dre_taster = LOW;
int dre_StateIncr = 1; //Status des Frequenz-Inkrements
unsigned char dre_TastZust =0;
unsigned long dre_timeT = 0;
unsigned long dre_akt_time = 0;
volatile char dre_p;// Flag Taster Drehgeber Flanke erkannt
volatile unsigned char dre_geber; // Geberwert nach Interrupt
unsigned char dre_val; //binaerer Geberwert
unsigned char dre_val_alt; //binaerer Geberwert letzter Durchlauf
//******************************************************************************
//EEPROM
int EE_Tast = 5;// Eingang fuer EE-Taster


// Definition einer Struktur-Variablen
// Ablagestruktur fuer EE_Ablage (frequenz / Inkrement)
struct EE_Konfig
{
  unsigned long f; //Ablage Startfrequenz 
  unsigned long s; // Ablage Stufe Frequenz-Inkrement
  byte check; // Ablage Kennzeichnung Gueltigkeit
}EE_Ablage;

//******************************************************************************
//Analogeingang, Pegelanzeige
int peg_anaPin = 0;// Analogeingang auf pin A0
int peg_ana_Wert = 0;
int peg_wert = 0;
int peg_ana_Wert_alt = 1;
int peg_rest = 0;
unsigned int pos_max;
char pos; // Index fuer Pegel-Cursor
#define PEG_BREIT 6
#define PEG_START 10
#define PEG_STUFE (1023/(PEG_BREIT+1))
//******************************************************************************
// LCD-Anzeige

// Connections:
// Analogeingang Arduino an pin A0
// Encoder Arduino pin 2, pin3
// Taster Encoder Arduino pin 4
// Taster EEPROM Arduino pin 5
// Arduino analog input 5 - I2C SCL
// Arduino analog input 4 - I2C SDA
// rs (LCD pin 4) to Arduino pin 12
// rw (LCD pin 5) to Arduino pin 11
// enable (LCD pin 6) to Arduino pin 10
// LCD pins d4, d5, d6, d7 to Arduino pins 9, 8, 7, 6
LiquidCrystal lcd(12, 11, 10, 9, 8, 7, 6);

// Definition eigener Symbole zur Pegeldarstellung und Inkrem.stufe
// nach http://www.quinapalus.com/hd44780udg.html
//
byte n1[8] = {
    0x0,0x0,0x0,0x0,0x0,0x0,0x15}; // Punkte 0-Limie
byte n2[8] = {
  0x10,0x10,0x10,0x10,0x10,0x10,0x15}; // 1 Linie
byte n3[8] = {
  0x14,0x14,0x14,0x14,0x14,0x14,0x15}; // 2 Linien
byte n4[8] = {
  0x15,0x15,0x15,0x15,0x15,0x15,0x15}; //3 Linien
byte n5[8] = {
  0x0,0x0,0x0f,0x0,0x0,0x0,0x0};
byte n6[8] = {
  0x0,0x0,0x0f,0x0,0x0,0x0,0x0};
byte n7[8] = {
  0x0,0x0,0x0f,0x0,0x0,0x0,0x0};

byte n8[8] = {
  0x0,0xe,0x11,0x15,0x11,0xe,0x0}; //Kreis mit Punkt


void setup()
{

  Serial.begin(9600);

  //Interrupsystem einrichten
  pinMode(dre_IntPin2, INPUT);
  pinMode(dre_IntPin3, INPUT);  
  
  attachInterrupt(0, GeberImpuls, CHANGE); //Pin 2 
  attachInterrupt(1, GeberImpuls, CHANGE); // Pin 3

  //Tastschalter Drehgeber;
  pinMode(dre_Tast, INPUT); 
  dre_p = 0;
  dre_val_alt = 0;

  //Taster EEPROM
  pinMode(EE_Tast,INPUT); 

  //Initialisierung der LCD-Anzeige
  // columns, rows.  use 16,2 for a 16x2 LCD, etc.
  lcd.begin(16,2);             
  //Kundenspez. Char. in Anzeige laden

  //Ablage der Zeichen von 1 bis 8
  lcd.createChar(1, n1); // Lasden der kundenspez. Character in Anzeige-Char.tabelle
  delay(30);
  lcd.createChar(2, n2);  
  delay(30);  
  lcd.createChar(3, n3);  
  delay(30);  
  lcd.createChar(4, n4);
  delay(30);  
  lcd.createChar(5, n5);
  delay(300);  
  lcd.createChar(6, n6);  
  delay(30);  
  lcd.createChar(7, n7);
  delay(30);
  lcd.createChar(8, n8);
  delay(30);

   // Initialisierungstext ´Display ausgeben  
  lcd.clear();             
  delay(100);
  lcd.setCursor(0,0);          
  lcd.print("Si5351-VFO V10");    
  lcd.setCursor(0,1);    
  delay(30);
  lcd.print("de DK4AQ, ");
  delay(1000);
 
  //Konfiguration aus EEPROM lesen (Startfreq., Inkrement)
  loadEEconfig();

  //Konfiguration Frequenzeinstellung initialisieren (Startfreq., Inkrement)
  initFreConfig();

  lcd.setCursor(0,0);  
  dis_ccc = (int)lcd.print(fre_frequenz); 

    // Initialisierung des Si5351
  si5351.init(SI5351_CRYSTAL_LOAD_8PF, 0);
  // Set CLK0 to output EDE-config. start freq.
  si5351.set_pll(SI5351_PLL_FIXED, SI5351_PLLA);
  si5351.set_freq((long long unsigned int)fre_frequenz * 100ULL, SI5351_PLL_FIXED, SI5351_CLK0);
 
  
  lcd.setCursor(12,0);
  lcd.print("Peg."); //Beschriftung Pegel ausgeben
  //Punktreihe Pegel.-Balken ausgeben
  for(pos = PEG_START; pos <= (PEG_START+PEG_BREIT) ;pos++) 
  {
    lcd.setCursor(pos,1);
    lcd.write(1); 
  }

  lcd.setCursor(7,0); //Beschriftung Frequenzwert 
  lcd.print(" Hz");
  delay(1000);  

  //Increment anzeigen
  dispIncrem();

}

void loop()
{

  char dis_m= 0; // Zaehlvariable Leer-Character

  // Bearbeitung der Meldung Drehgeber Pulse (Flag p),
  // Veraenderung der Sollfrequenz
  while( dre_p > 0)
    {
      dre_val = dre_value[dre_geber];
  
      char muster = dre_val - dre_val_alt;
      
      // Ausgangsmuster der Opto-Encoders
      //              Output A    Output B
      // Position 0:    0           0
      // Position 1:    1           0
      // Position 2:    1           1
      // Position 3:    0           1
      
      //Erkennung der Musteruebergaenge des Drehgebers 
      switch(muster)
      {
        case 1:
        case 2:
            if(fre_increm_neu == HIGH)
              {fre_frequenz = fre_frequenz /fre_deltaF;
               fre_frequenz = fre_frequenz *fre_deltaF;
               fre_increm_neu = LOW;
              }
            fre_frequenz = fre_frequenz + fre_deltaF;
 
            break;
        case 3:
            if(fre_increm_neu == HIGH)
              {fre_frequenz = fre_frequenz /fre_deltaF;
               fre_frequenz = fre_frequenz *fre_deltaF;
               fre_increm_neu = LOW;
              }
            else
              {
                fre_frequenz = fre_frequenz - fre_deltaF;
              }
         
            break;  
        case -1:
        case -2:
            if(fre_increm_neu == HIGH)
              {fre_frequenz = fre_frequenz /fre_deltaF;
               fre_frequenz = fre_frequenz *fre_deltaF;
               fre_increm_neu = LOW;
              }
            else
              {
                fre_frequenz = fre_frequenz - fre_deltaF;
              }
       
            break; 
        case -3:
             if(fre_increm_neu == HIGH)
              {fre_frequenz = fre_frequenz /fre_deltaF;
               fre_frequenz = fre_frequenz *fre_deltaF;
               fre_increm_neu = LOW;
               }
            fre_frequenz = fre_frequenz + fre_deltaF;
            //Serial.print("+")
            break;
        default:
            break;
      }  
      
      dre_p = 0;
      dre_val_alt = dre_val;
  
  }

  // Aktualisierung der Sollfrequenz
  if((fre_frequenz != fre_frequenz_alt)&&(dre_p == 0))
  {
    // Begrenzung Anzeigebereich
    if(fre_frequenz > fre_max) fre_frequenz = fre_max; 
    if(fre_frequenz < fre_min) fre_frequenz = fre_min;
    
    //Display Frequenz
    lcd.setCursor(0,0);  
    dis_ccc = (int)lcd.print(fre_frequenz); 
   
    //Ueberschreiben der nicht geschriebenen Felder mit Leerzeichen  
    for(dis_m= (7 - dis_ccc); dis_m>0; dis_m--)
    {
      lcd.print(" ");
    }  
   
   
    // Ausgabe Frequenz an Si5351
    si5351.set_pll(SI5351_PLL_FIXED, SI5351_PLLA);
    //Casting der 4-Byte-Long Variablen auf eine 8-Byte Long-Variable
    // Uebergabe der Sollfrequenz
    si5351.set_freq((long long unsigned int)fre_frequenz * 100ULL, SI5351_PLL_FIXED, SI5351_CLK0);
    fre_frequenz_alt = fre_frequenz;
    fre_new = LOW;
  }
  else
  {
    // Ausgabe Pegelwert
    // Schreiben der Pegel-Symbole in Anzeige
    dispPegel();
  }
   
  // Umschaltung und Anzeige des Frequenz-Inkrements durch den Drehgeber-Taster
  chngIncrem();
  
  //Abspeicherung des Startwertes Frequenz und Frequenz-Inkrement
  //ins EEPROM durch Bedientaste. 

  chngConfig();
}
  
// Interrupt Routine (Variante Opto-Drehgeber)
void GeberImpuls()
{
   dre_geber = (char)digitalRead(dre_IntPin2) << 1;
   dre_geber = dre_geber | (char)digitalRead(dre_IntPin3);

   dre_p++;
  //noInterrupts();
}














